/*
** Module   : file_system_helper_linux.c
** Abstract : implementation for detection of simbolic links, non-FIFO pipes and special devices.
**            This is used by filelys.c
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130329 Created initial version to implement Linux specifics for file system
**                  operations.
*/
#include "filesys.h"

/**
* The helper function to get symlink status in Linux.
*
* @param  env 
*         pointer to a structure that provides access to JNI environment.
* @param  fname 
*         The file name to be checked passed from FileChecker.checkFileType().
* @param  file_path 
*         The pointer to file path string.
* @param  status_initial 
*         The current status value for the given file.
*
* @return An integer modified status code of the file.
*/
int getSymLinkStatus(JNIEnv *env, jstring fname, const char *file_path, int status_initial)
{
   // Start with current version
   int status_return = status_initial;

   struct stat buf;
   //get symbolic link status
   //we can't use fstat for detecting symbolic links
   
   int lnk_stat = lstat(file_path, &buf); 
   
   (*env)->ReleaseStringUTFChars(env, fname, file_path);
   //if link status is -1 return -1
   if (lnk_stat==-1) return -1;

   
   // check for special devices
   if (S_ISCHR(buf.st_mode) || S_ISBLK(buf.st_mode))
   { 
      // status code for devices is standalone
      status_return |= DEVICE;
   }
   // check for symbolic links
   // NOTE: A symbolic link that points to a directory or file will not be
   // detected as file or directory by the logic present here so this status 
   //cannot be made inclusive with file or directory
   if (S_ISLNK(buf.st_mode))
   {
      //status code for symbolic link is additive
      status_return |= SYMLINK;
   }
   
   return status_return;
}

/**
* The helper function to open file or pipe in Linux.
*
* @param  filename 
*         The name of the file to open.
* @param  flags 
*         The open mode flags common across all operating systems.
*
* @return An integer file fescriptor or -1 in case of error.
*/
int open_native(const char * filename, int flags)
{
   return open(filename, flags | O_NONBLOCK);
}
